# Standard Imports
import threading
from typing import Any, ClassVar

# Project Imports

# Third-party imports


class SingletonMeta[T](type):
    _instances: ClassVar[dict] = {}

    _lock: threading.Lock = threading.Lock()

    def __call__(cls, *args: Any, **kwargs: Any) -> T:  #  noqa: ANN401
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]



# Standard Imports
from __future__ import annotations

import logging
import uuid
from collections.abc import Awaitable, Callable  # noqa: TC003 False Positive
from functools import wraps
from typing import TYPE_CHECKING, Any, ClassVar, Self

# Third Party Imports
from pydantic.dataclasses import dataclass

# Project Imports
from common.type_def import ETLData  # noqa: TC001 False Positive
from common.utils import SingletonMeta

if TYPE_CHECKING:
    from common.type_def import PluginName, WrappedPlugin
    from core.models.phases import PipelinePhase


def plugin(plugin_phase: PipelinePhase, plugin_name: str) -> Callable[[WrappedPlugin], WrappedPlugin]:
    @wraps(plugin)
    def decorator(func: WrappedPlugin) -> WrappedPlugin:
        PluginRegistry.register(plugin_phase, plugin_name, func)
        return func

    return decorator


@dataclass
class PluginWrapper:
    id: str
    func: Callable[..., ETLData] | Callable[..., Awaitable[ETLData]]

    def __eq__(self: Self, other: object) -> bool:
        if not isinstance(other, PluginWrapper):
            return False
        return self.id == other.id and self.func.__name__ == other.func.__name__


class PluginRegistry(metaclass=SingletonMeta):
    """A Plugin Factory that dynamically registers, removes and fetches plugins.

    Registry example: {EXTRACT_PHASE: {'s3': S3Plugin}}
    """

    _registry: ClassVar[dict[PipelinePhase, dict[PluginName, WrappedPlugin]]] = {}

    @classmethod
    def register(
        cls,
        pipeline_phase: PipelinePhase,
        plugin_name: PluginName,
        plugin_callable: WrappedPlugin,
    ) -> bool:
        """Regisers a plugin for a given pipeline type and plugin."""
        # Initialise the Pipeline phase in the registry.
        if pipeline_phase not in cls._registry:
            cls._registry[pipeline_phase] = {}

        # Check if the plugin has been registered.
        if plugin_name in cls._registry[pipeline_phase]:
            logging.warning(
                "Plugin for `%s` phase already exists in PluginRegistry class.",
                pipeline_phase,
            )
            return False

        cls._registry[pipeline_phase][plugin_name] = plugin_callable
        logging.debug(
            "Plugin `%s` for phase `%s` registered successfully.",
            plugin,
            pipeline_phase,
        )
        return True

    @classmethod
    def remove(cls, pipeline_phase: PipelinePhase, plugin_name: PluginName) -> bool:
        """Remove a plugin for a given pipeline type and plugin."""
        if pipeline_phase in cls._registry and plugin_name in cls._registry[pipeline_phase]:
            del cls._registry[pipeline_phase][plugin_name]
            logging.debug("Plugin '%s' for phase '%s' has been removed.", plugin_name, pipeline_phase)

            # Remove the ETL type dict if empty after removing the plugin.
            if not cls._registry[pipeline_phase]:
                logging.debug("Stage '%s' has been removed.", pipeline_phase)
                del cls._registry[pipeline_phase]
            return True
        return False

    @classmethod
    def get(cls, pipeline_phase: PipelinePhase, plugin_name: PluginName) -> WrappedPlugin:
        """Retrieve a plugin for a given pipeline type and plugin."""
        plugin_factory = cls._registry.get(pipeline_phase, {}).get(plugin_name, None)

        if not plugin_factory:
            msg = f"Plugin class was not found for following plugin `{plugin_name}`."
            raise ValueError(msg)

        logging.debug("Plugin class '%s' has been successfully retrieved.", plugin_factory)
        return plugin_factory

    @classmethod
    def instantiate_plugin(cls, phase_pipeline: PipelinePhase, plugin_data: dict[str, Any]) -> PluginWrapper:
        """Resolve and return a single plugin instance."""
        plugin_name = plugin_data.pop("plugin", None)
        if not plugin_name:
            raise ValueError("The attribute 'plugin' is empty.")

        plugin_factory = cls.get(phase_pipeline, plugin_name)

        plugin_id = plugin_data.pop("id", None) or f"{plugin_factory.__name__}_{uuid.uuid4()}"
        plugin_params = plugin_data.get("params", {})
        if plugin_params:
            return PluginWrapper(id=plugin_id, func=plugin_factory(**plugin_params))

        return PluginWrapper(id=plugin_id, func=plugin_factory())



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Self

import aiofiles
import yaml
import yamlcore

if TYPE_CHECKING:
    from common.type_def import PluginRegistryJSON

from pydantic.dataclasses import dataclass

from common.utils import SingletonMeta

type JSON_DATA = dict

DEFAULT_CONCURRENCY = 2
DEFAULT_ENGINE = "native"


class YamlConfig(metaclass=SingletonMeta):
    engine: str
    concurrency: int

    def __init__(self, engine: str = DEFAULT_ENGINE, concurrency: int = DEFAULT_CONCURRENCY) -> None:
        self.engine = engine
        self.concurrency = concurrency


class YamlAttribute(Enum):
    PIPELINES = "pipelines"
    PLUGINS = "plugins"
    ENGINE = "engine"
    CONCURRENCY = "concurrency"


@dataclass
class YamlParser:
    content: dict

    @classmethod
    def from_text(cls, yaml_text: str) -> YamlParser:
        return cls(yaml.load(yaml_text, Loader=yamlcore.CoreLoader))  # noqa: S506 - Extension of PyYAML YAML 1.2 Compliant

    @classmethod
    async def from_file(cls, file_path: str, encoding: str = "utf-8") -> YamlParser:
        """Create YamlDocument from file."""
        try:
            async with aiofiles.open(file_path, encoding=encoding) as file:
                content = await file.read()
                return cls.from_text(content)

        except FileNotFoundError as error:
            error_msg = f"File not found: {file_path}"
            raise FileNotFoundError(error_msg) from error

    def get_pipelines_dict(self: Self) -> JSON_DATA:
        """Return the 'pipelines' section from the parsed YAML."""
        return self.content.get(YamlAttribute.PIPELINES.value, {})

    def get_plugins_dict(self: Self) -> PluginRegistryJSON | None:
        return self.content.get(YamlAttribute.PLUGINS.value, None)

    def initialize_yaml_config(self: Self) -> YamlConfig:
        # Create the map of attributes with their values
        attrs_map = {
            YamlAttribute.ENGINE.value: self.content.get(YamlAttribute.ENGINE.value, DEFAULT_ENGINE),
            YamlAttribute.CONCURRENCY.value: self.content.get(YamlAttribute.CONCURRENCY.value, DEFAULT_CONCURRENCY),
        }

        # Filter out the None values
        attrs = {key: value for key, value in attrs_map.items() if value is not None}

        return YamlConfig(**attrs)



from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Annotated, Any, TypedDict

from pydantic.dataclasses import dataclass

type ExtractedData = Any
type TransformedData = Any

type ETLData = ExtractedData | TransformedData
type PluginName = str

type SyncPlugin = Callable[..., ETLData]
type AsyncPlugin = Callable[..., Awaitable[ETLData]]
type WrappedPlugin[**PluginArgs] = Callable[PluginArgs, SyncPlugin] | Callable[PluginArgs, AsyncPlugin]


class CustomPluginRegistryJSON(TypedDict):
    dirs: Annotated[list[str], "List of directories to dynamically import for custom plugins"]
    files: Annotated[list[str], "List of files to dynamically import for custom plugins"]


class PluginRegistryJSON(TypedDict):
    custom: CustomPluginRegistryJSON
    community: Annotated[list[str], "List of community plugins to import"]


@dataclass
class ExtractStageResult:
    id: str
    success: bool
    data: ExtractedData
    error: str | None = None


@dataclass
class TransformStageResult:
    id: str
    success: bool
    data: TransformedData
    error: str | None = None


@dataclass
class LoadStageResult:
    id: str
    success: bool
    error: str | None = None


@dataclass
class TransformLoadStageResult:
    id: str
    success: bool
    error: str | None = None



# Standard Imports
import logging

from common.logger import setup_logger
from core.loaders import load_plugins
from core.orchestrator import PipelineOrchestrator

# # Project Imports
from core.parsers import YamlParser, parse_pipelines


async def start(yaml_text: str | None = None, file_path: str | None = None) -> bool:
    # Set up the logger configuration
    setup_logger()

    # Parse YAML and initialize YAML config (engine, concurrency)
    if yaml_text is None and file_path is None:
        raise ValueError("Either yaml_text or file_path must be provided.")

    yaml_parser = (
        await YamlParser.from_file(file_path) if file_path else YamlParser.from_text(yaml_text) if yaml_text else None
    )
    if not yaml_parser:
        raise ValueError("YamlParser could not be initialized.")

    yaml_config = yaml_parser.initialize_yaml_config()
    plugins_payload = yaml_parser.get_plugins_dict()

    # Parse plugins directly within the load_plugins function
    load_plugins(yaml_config.engine, plugins_payload)

    # Parse pipelines and execute them using the orchestrator
    pipelines = parse_pipelines(yaml_parser.get_pipelines_dict())

    try:
        orchestrator = PipelineOrchestrator(yaml_config)
        await orchestrator.execute_pipelines(pipelines)

    except Exception as e:
        logging.error("The following error occurred: %s", e)
        logging.error("The original cause is: %s", e.__cause__)
        raise
    else:
        return True



# Standard Imports
from __future__ import annotations

import logging
from enum import Enum, unique
from typing import TYPE_CHECKING, Annotated, Self

if TYPE_CHECKING:
    from collections.abc import Callable

# Third-party Imports
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    model_validator,
)

# Project Imports
from core.plugins import PluginRegistry, PluginWrapper

# A callable class type representing all phase objects.
type Phase = ExtractPhase | TransformPhase | LoadPhase | TransformLoadPhase


@unique
class PipelinePhase(Enum):
    EXTRACT_PHASE = "extract"
    LOAD_PHASE = "load"
    TRANSFORM_PHASE = "transform"
    TRANSFORM_AT_LOAD_PHASE = "transform_at_load"


def serialize_plugin(phase_pipeline: PipelinePhase) -> Callable[[dict], PluginWrapper]:
    def validator(value: dict) -> PluginWrapper:
        return PluginRegistry.instantiate_plugin(phase_pipeline, value)

    return validator


def serialize_plugins(phase_pipeline: PipelinePhase) -> Callable[[list], list[PluginWrapper]]:
    def validator(value: list) -> list[PluginWrapper]:
        return [PluginRegistry.instantiate_plugin(phase_pipeline, plugin_dict) for plugin_dict in value]

    return validator


def unique_id_validator(steps: list[PluginWrapper]) -> list[PluginWrapper]:
    if not steps:
        logging.debug("No plugins to validate for unique ID.")
        return steps

    ids = {}

    for step in steps:
        if step.id in ids:
            raise ValueError("The `ID` is not unique. There already exists an 'ID' with this name.")

        ids[step.id] = 1

    return steps


class ExtractPhase(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    steps: Annotated[
        list[PluginWrapper],
        Field(min_length=1),
        BeforeValidator(serialize_plugins(PipelinePhase.EXTRACT_PHASE)),
        AfterValidator(unique_id_validator),
    ]
    pre: Annotated[
        list[PluginWrapper] | None,
        BeforeValidator(serialize_plugins(PipelinePhase.EXTRACT_PHASE)),
    ] = None

    merge: Annotated[
        PluginWrapper | None,
        BeforeValidator(serialize_plugin(PipelinePhase.EXTRACT_PHASE)),
    ] = None

    @model_validator(mode="after")
    def check_merge_condition(self: Self) -> Self:
        if self.merge and not len(self.steps) > 1:
            raise ValueError("Validation Error! Merge can only be used if there is more than one extract step.")

        if len(self.steps) > 1 and not self.merge:
            raise ValueError("Validation Error! Merge is required when there are more than one extract step.")

        return self


class TransformPhase(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    steps: Annotated[
        list[PluginWrapper],
        BeforeValidator(serialize_plugins(PipelinePhase.TRANSFORM_PHASE)),
    ]


class LoadPhase(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    steps: Annotated[
        list[PluginWrapper],
        Field(min_length=1),
        BeforeValidator(serialize_plugins(PipelinePhase.LOAD_PHASE)),
    ]
    pre: Annotated[
        list[PluginWrapper] | None,
        BeforeValidator(serialize_plugins(PipelinePhase.LOAD_PHASE)),
    ] = None

    post: Annotated[
        list[PluginWrapper] | None,
        BeforeValidator(serialize_plugins(PipelinePhase.LOAD_PHASE)),
    ] = None


class TransformLoadPhase(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    steps: Annotated[
        list[PluginWrapper],
        Field(min_length=1),
        BeforeValidator(serialize_plugins(PipelinePhase.TRANSFORM_AT_LOAD_PHASE)),
    ]
